import urllib3
from reho.model.preprocessing.QBuildings import *
import time
from datetime import date, datetime
from typing import Dict, Optional, Union
import requests as rq
from shapely import wkt
import json

FIND_PATTERN = "PREFIX\\s*(.*?)\n"
SPLIT_PATTERN = ":\\s*"

DATA_TYPES_TO_PYTHON_CLS = {
    "http://www.w3.org/2001/XMLSchema#integer": int,
    "http://www.w3.org/2001/XMLSchema#float": float,
    "http://www.w3.org/2001/XMLSchema#double": float,
    "http://www.w3.org/2001/XMLSchema#decimal": float,
    "http://www.w3.org/2001/XMLSchema#date": date.fromisoformat,
    "http://www.w3.org/2001/XMLSchema#dateTime": (
        lambda x: datetime.strptime(x.rstrip("Z"), "%Y-%m-%dT%H:%M:%S")
    ),
    "https://www.opengis.net/ont/geosparql#wktLiteral": wkt.loads,
    "https://www.openlinksw.com/schemas/virtrdf#Geometry": wkt.loads,
    "https://www.w3.org/2001/XMLSchema#integer": int,
    "https://www.w3.org/2001/XMLSchema#float": float,
    "https://www.w3.org/2001/XMLSchema#double": float,
    "https://www.w3.org/2001/XMLSchema#decimal": float,
    "https://www.w3.org/2001/XMLSchema#date": date.fromisoformat,
    "https://www.w3.org/2001/XMLSchema#dateTime": (
        lambda x: datetime.strptime(x.rstrip("Z"), "%Y-%m-%dT%H:%M:%S")
    ),
}

GEODATA_TYPES = {"https://www.opengis.net/ont/geosparql#wktLiteral", "https://www.openlinksw.com/schemas/virtrdf#Geometry",
                 "https://www.opengis.net/ont/geosparql#wktLiteral", "https://www.openlinksw.com/schemas/virtrdf#Geometry"}


def requests_retry_session(
        retries=3, backoff_factor=0.3, status_forcelist=(500, 502, 504), session=None
):
    session = session or rq.Session()
    retry = urllib3.util.Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = rq.adapters.HTTPAdapter(max_retries=retry)
    session.mount("https://", adapter)
    session.mount("https://", adapter)

    return session


class SparqlClient:
    def __init__(
            self,
            base_url: str = None,
            timeout: Optional[int] = 15,
            output: Optional[str] = "pandas",
            user: Optional[str] = None,
            password: Optional[str] = None,
    ) -> None:

        self.BASE_URL = base_url
        self.last_request = 0
        self.HEADERS = {
            "Accept": "application/sparql-results+json",
        }
        self.prefixes = dict()
        self.timeout = timeout
        self.output = output

        self.session = requests_retry_session()
        if user and password:
            self.session.auth = (user, password)

    def _normalize_prefixes(self, prefixes: Dict) -> str:
        """Transfrom prefixes map to SPARQL-readable format
        Args:
            prefixes: 		prefixes to be normalized

        Returns
            str             SPARQL-readable prefix definition
        """

        normalized_prefixes = "\n".join(
            "PREFIX %s" % ": ".join(map(str, x)) for x in prefixes.items()
        )
        if normalized_prefixes:
            normalized_prefixes += "\n"
        return normalized_prefixes

    def add_prefixes(self, prefixes: Dict) -> None:
        """Define prefixes to be added to every query
        Args:
            prefixes: prefixes to be added to every query

        Returns
            None
        """

        self.prefixes = {**self.prefixes, **prefixes}

    def remove_prefixes(self, prefixes: Dict) -> None:
        """Remove prefixes from the prefixes are added to every query
        Args:
            prefixes: prefixes to be removed from self.prefixes

        Returns
            None
        """

        for prefix in prefixes:
            self.prefixes.pop(prefix, None)

    def _format_query(self, query: str) -> str:
        """Format SPARQL query to include in-memory prefixes.
        Prefixes already defined in the query have precedence, and are not overwritten.
            Args:
                query: user-defined SPARQL query

            Returns
                str: SPARQL query with predefined prefixes
        """

        prefixes_in_query = dict(
            [
                re.split(SPLIT_PATTERN, prefix, 1)
                for prefix in re.findall(FIND_PATTERN, query)
            ]
        )
        prefixes_to_add = {
            k: v for (k, v) in self.prefixes.items() if k not in prefixes_in_query
        }

        return self._normalize_prefixes(prefixes_to_add) + query

    def send_query(self, query: str, timeout: Optional[int] = 0) -> pd.DataFrame:
        """Send SPARQL query. Transform results to pd.DataFrame.
        Args:
            query: 				full SPARQL query
            timeout:            timeout (in seconds) for this query. If not defined, the self.timeout will be used.

        Returns
            pd.DataFrame	    query results
        """

        request = {"query": self._format_query(query)}

        if time.time() < self.last_request + 1:
            time.sleep(1)
        self.last_request = time.time()

        if timeout == 0:
            timeout = self.timeout

        if timeout is not None:
            response = self.session.get(
                self.BASE_URL, headers=self.HEADERS, params=request, timeout=timeout
            )
        else:
            response = self.session.get(
                self.BASE_URL, headers=self.HEADERS, params=request
            )
        response.raise_for_status()
        response = response.json()

        if "head" not in response:
            raise ExecutionError(
                "{}\n Triplestore error code: {}".format(
                    response["message"], response["code"]
                )
            )

        if not response["results"]["bindings"]:
            raise NotFoundError()

        if self.output == "pandas":
            return self._normalize_results(response)
        elif self.output == "dict":
            return response
        else:
            raise TypeError(
                "Invalid output type. Choose `pandas` or `dict` as output type"
            )

    def _normalize_results(
            self, response: Dict
    ) -> Union[pd.DataFrame, gpd.GeoDataFrame]:
        """Normalize response from SPARQL endpoint. Transform json structure to table. Convert observations to python data types.
        Args:
            response: 			raw response from SPARQL endpoint

        Returns
            pd.DataFrame: response from SPARQL endpoint in a tabular form, with python data types
        """

        cols = response["head"]["vars"]
        data = dict(zip(cols, [[] for i in range(len(cols))]))

        has_geo_data = False
        for row in response["results"]["bindings"]:
            for key in cols:

                if key in row:

                    value = row[key]["value"]
                    if "datatype" in row[key]:
                        datatype = row[key]["datatype"]
                        value = DATA_TYPES_TO_PYTHON_CLS[datatype](value)

                        if datatype in GEODATA_TYPES:
                            has_geo_data = True
                            geometry_col = key

                else:
                    value = None

                data[key].append(value)

        if has_geo_data:
            return gpd.GeoDataFrame.from_dict(data).set_geometry(col=geometry_col)
        else:
            return pd.DataFrame.from_dict(data)


class GraphlyError(Exception):
    pass


class NotFoundError(GraphlyError):
    pass


class ExecutionError(GraphlyError):
    pass


cantons = {'id': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26],
           'name_fr': ['Zurich', 'Berne', 'Lucerne', 'Uri', 'Schwytz', 'Obwald', 'Nidwald', 'Glaris', 'Zoug',
                       'Fribourg',
                       'Soleure', 'Bâle-ville', 'Bâle-campagne', 'Schaffhouse',
                       'Appenzell Rhodes-Extérieures', 'Appenzell Rhodes-Intérieures', 'St-Gall', 'Grisons', 'Argovie',
                       'Thurgovie', 'Tessin', 'Vaud', 'Valais', 'Neuchâtel', 'Genève', 'Jura'],
           'name_de': ["Zürich", 'Bern', 'Luzern', 'Uri', "Schwyz", "Obwalden", "Nidwalden", "Glarus", "Zug",
                       'Freiburg',
                       "Solothurn", "Basel-Stadt", 'Basel-Landschaft', "Schaffhausen",
                       'Appenzell Ausserrhoden', 'Appenzell Innerrhoden', 'St. Gallen', "Graubünden", "Aargau",
                       'Thurgau', 'Tessin', 'Waadt', 'Wallis', 'Neuenburg', 'Genf', 'Jura'],
           'name_en': ["Zurich", 'Bern', 'Lucerne', 'Uri', "Schwyz", "Obwalden", "Nidwalden", "Glarus", "Zug",
                       'Freiburg',
                       "Solothurn", "Basel-City", 'Basel-Country', "Schaffhausen",
                       'Appenzell Outer-Rhodes', 'Appenzell Inner-Rhodes', 'St. Gallen', "Graubünden", "Aargau",
                       'Thurgau', 'Tessin', 'Vaud', 'Valais', 'Neuchâtel', 'Geneva', 'Jura']}
cantons = pd.DataFrame.from_dict(cantons).set_index('id')

sparql = SparqlClient("https://lindas.admin.ch/query")
geosparql = SparqlClient("https://ld.geo.admin.ch/query")

sparql.add_prefixes({
    "schema": "<https://schema.org/>",
    "cube": "<https://cube.link/>",
    "elcom": "<https://energy.ld.admin.ch/elcom/electricityprice/dimension/>",
    "admin": "<https://schema.ld.admin.ch/>",
})

geosparql.add_prefixes({
    "dct": "<https://purl.org/dc/terms/>",
    "geonames": "<https://www.geonames.org/ontology#>",
    "schema": "<https://schema.org/>",
    "geosparql": "<https://www.opengis.net/ont/geosparql#>",
})


def get_providers_by_municipality_id(city=None, from_csv=False):
    """
    Gives the electricity providers for a given municipality.

    Parameters
    ----------
    city : (int, str, QBuildingsReader)
        The city from which the providers should be retrieved. It can be in the form of its ID, its name or retrieved by
        the QBuildingsReader
        If None is passed, the whole correspondance table between the municipalities and the providers is returned.
    from_csv : bool
        A correspondance table has been saved as a csv to avoid querying the ELCOM database. However, if you want to
        be sure that it is the last municipality IDs or if you know some providers have changed since the date of the
        corresponding table, set it to False.

    Returns
    -------
    pd.DataFrame which columns are ['id_city', 'commune', 'id_operator', 'operator'] for the given city.
    Notes
    -----
    - The correspondance table is dated from 01.02.2024
    - ``QBuildingsReader`` object can be passed to 'city' for automatic localization, from the initial building selection.

    Examples
    --------
    >>> get_providers_by_municipality_id(city='Rüderswil')
    """
    city_link = 'https://ld.admin.ch/municipality/'
    op_link = 'https://energy.ld.admin.ch/elcom/electricityprice/operator/'

    if isinstance(city, QBuildingsReader):
        city = city.data['transformers']['id_city'][0]
        city_query = '<' + city_link + str(city) + '>'
    elif isinstance(city, int):
        city_query = '<' + city_link + str(city) + '>'
    else:
        city_query = "?id_city"

    #  Use csv of correspondance from 01.02.2024
    if from_csv:
        communes = pd.read_csv(os.path.join(path_to_elcom, 'correspondance_table_municipality_operator.csv'),
                               index_col=0)
        if city is not None:
            mask = (communes['id_city'] == city) + (communes['commune'] == city)
            communes = communes.loc[mask]
            if communes.empty:
                print('No corresponding city to the given identifier')

    else:
        query = """
            SELECT DISTINCT ?id_city ?commune ?id_operator ?operator WHERE {
                <https://energy.ld.admin.ch/elcom/electricityprice> <https://cube.link/observationSet> ?observationSet0 .
                ?observationSet0 <https://cube.link/observation> ?source0 .
            ?source0 <https://energy.ld.admin.ch/elcom/electricityprice/dimension/municipality> city_to_replace .
            ?source0 <https://energy.ld.admin.ch/elcom/electricityprice/dimension/operator> ?id_operator .
            
            OPTIONAL {
                city_to_replace <https://schema.org/name> ?commune_0 .
                FILTER (
                  LANGMATCHES(LANG(?commune_0), "fr")
                )
              }
              OPTIONAL {
                city_to_replace <https://schema.org/name> ?commune_1 .
                FILTER (
                  LANGMATCHES(LANG(?commune_1), "de")
                )
              }
              OPTIONAL {
                city_to_replace <https://schema.org/name> ?commune_2 .
                FILTER (
                  LANGMATCHES(LANG(?commune_2), "it")
                )
              }
              OPTIONAL {
                city_to_replace <https://schema.org/name> ?commune_3 .
                FILTER (
                  LANGMATCHES(LANG(?commune_3), "en")
                )
              }
              OPTIONAL {
                city_to_replace <https://schema.org/name> ?commune_4 .
                FILTER (
                  (LANG(?commune_4) = "")
                )
              }
            BIND(COALESCE(?commune_0, ?commune_1, ?commune_2, ?commune_3, ?commune_4) AS ?commune)    
            
            OPTIONAL {
                ?id_operator <https://schema.org/name> ?operator_0 .
                FILTER (
                  LANGMATCHES(LANG(?operator_0), "fr")
                )
              }
              OPTIONAL {
                ?id_operator <https://schema.org/name> ?operator_1 .
                FILTER (
                  LANGMATCHES(LANG(?operator_1), "de")
                )
              }
              OPTIONAL {
                ?id_operator <https://schema.org/name> ?operator_2 .
                FILTER (
                  LANGMATCHES(LANG(?operator_2), "it")
                )
              }
              OPTIONAL {
                ?id_operator <https://schema.org/name> ?operator_3 .
                FILTER (
                  LANGMATCHES(LANG(?operator_3), "en")
                )
              }
              OPTIONAL {
                ?id_operator <https://schema.org/name> ?operator_4 .
                FILTER (
                  (LANG(?operator_4) = "")
                )
              }
              BIND(COALESCE(?operator_0, ?operator_1, ?operator_2, ?operator_3, ?operator_4) AS ?operator)
            }
            """
        query = query.replace('city_to_replace', city_query)
        communes = sparql.send_query(query)
        if communes.loc[0, 'id_city'] is None:
            communes['id_city'] = int(city)
        else:
            communes['id_city'] = communes['id_city'].apply(lambda row: int(row.split(city_link)[1]))
        communes['id_operator'] = communes['id_operator'].apply(lambda row: int(row.split(op_link)[1]))

    return communes


def get_prices_from_elcom_by_canton(year=2024, canton=None, category=None, tva=None, export_path=None):
    """
    Queries the electricity retail prices from the ELCOM database.
    Year, canton and consumer category can be specified.
    TVA is applied by default or can be adapted as a scaling factor.

    Parameters
    ----------
    year : int
        Year from which the electricity prices must be retrieved.
    canton : str/int
        Canton from which the electricity prices must be retrieved. Can be in form of canton ID or canton name.
    category : str
        Category from which the electricity prices must be retrieved.
    tva : bool
        Whether the tva should be included in the final results or not.
    export_path : str
        If given, export the prices with the parameter required at the path.

    Returns
    -------
    pd.DataFrame with the electricity price and its components.

    See also
    --------
    get_prices_from_elcom_by_city : To retrieve the ELCOM prices by city.
    get_injection_prices : To obtain the injection prices instead.

    Notes
    -----
    - A ``QBuildingsReader`` object can be passed to 'canton' for automatic localization.
    - List and description of the available categories are available at the `ELCOM website <https://www.prix-electricite.elcom.admin.ch/>`_.
    - The TVA on electricity changed in 2024, from 7.7% to 8.1%.

    Examples
    --------
    >>> prices = electricity_prices.get_prices_from_elcom_by_canton(canton='Geneva', category='H4')
    >>> prices
        Year  Canton Category  ...  community_fees  aidfee  finalcosts
    0  2024  Geneva       H4  ...         1.42824     2.3   30.925972
    [1 rows x 9 columns]
    """

    canton_link = 'https://ld.admin.ch/canton/'
    cat_link = 'https://energy.ld.admin.ch/elcom/electricityprice/category/'

    if isinstance(canton, QBuildingsReader):
        canton = canton.data['transformers']['id_canton'][0]
        canton_query = '<' + canton_link + str(canton) + '>'
        bonus_columns = ""
    elif isinstance(canton, str):
        mask = (cantons['name_fr'] == canton) + (cantons['name_de'] == canton) + (cantons['name_en'] == canton)
        canton = str(cantons[mask].first_valid_index())
        canton_query = '<' + canton_link + str(canton) + '>'
        bonus_columns = ""
    elif isinstance(canton, int):
        canton_query = '<' + canton_link + str(canton) + '>'
        bonus_columns = ""
    else:
        canton_query = "?Canton"
        bonus_columns = "?Canton"
    valid_cat = ['H1', 'H2', 'H3', 'H4', 'H5', 'H6', 'H7', 'H8', 'C1', 'C2', 'C3', 'C4', 'C5', 'C6', 'C7']
    if category in valid_cat:
        cat_query = '<' + cat_link + category + '>'
    elif category is None:
        cat_query = "?Category"
        bonus_columns += "?Category"
    else:
        raise ValueError("The category asked is not a valid one from the elcom.\n"
                         f"Please pick among {valid_cat}.\n")

    query_template = """
    SELECT ?Year ?Canton ?Category ?totalcosts ?energy ?grid ?community_fees ?aidfee
    WHERE {
        <https://energy.ld.admin.ch/elcom/electricityprice-canton> cube:observationSet ?observationSet0 .
        ?observationSet0 cube:observation ?observation .

        ?observation
          elcom:category category_to_replace;
          elcom:canton canton_to_replace;
          elcom:period "year_to_replace"^^<http://www.w3.org/2001/XMLSchema#gYear>;
          elcom:product <https://energy.ld.admin.ch/elcom/electricityprice/product/standard>;
          elcom:total ?totalcosts;
          elcom:gridusage ?grid;
          elcom:energy ?energy;
          elcom:charge ?community_fees;
          elcom:aidfee ?aidfee.
    }
    """

    query = query_template.replace('year_to_replace', str(year))
    query = query.replace('canton_to_replace', canton_query)
    query = query.replace('category_to_replace', str(cat_query))
    query = query.replace('add_columns', str(bonus_columns))
    prices = sparql.send_query(query)
    if prices.loc[0, 'Canton'] is None:
        prices['Canton'] = cantons.loc[int(canton), 'name_en']
    else:
        prices['Canton'] = prices['Canton'].apply(lambda row: cantons.loc[int(row.split(canton_link)[1]), 'name_en'])
    if prices.loc[0, 'Category'] is None:
        prices['Category'] = category
    else:
        prices['Category'] = prices['Category'].apply(lambda row: row.split(cat_link)[1])
    prices['Year'] = year

    # Define custom sorting order
    custom_order = {'H1': 1, 'H2': 2, 'H3': 3, 'H4': 4, 'H5': 5, 'H6': 6, 'H7': 7, 'H8': 8,
                    'C1': 9, 'C2': 10, 'C3': 11, 'C4': 12, 'C5': 13, 'C6': 14, 'C7': 15}
    prices['Category'] = prices['Category'].map(custom_order)

    # Sort the DataFrame based on the "Category" column
    prices = prices.sort_values(by='Category').reset_index(drop=True)

    # Convert back to original category format
    prices['Category'] = prices['Category'].map({v: k for k, v in custom_order.items()})

    if tva is None:
        if year >= 2024:
            tva = 1.081
        else:
            tva = 1.077

    prices['finalcosts'] = tva * prices['totalcosts']

    if isinstance(export_path, str):
        prices.to_csv(os.path.realpath(export_path))

    return prices


def get_prices_from_elcom_by_city(year=2024, city=None, category=None, tva=None, export_path=None):
    """
    Queries the electricity retail prices from the ELCOM database by munipalities.

    Year, municipality and consumer category can be specified.
    TVA is applied by default or can be adapted as a scaling factor.

    Parameters
    ----------
    year : int
        Year from which the electricity prices must be retrieved.
    city : str/int
        Municipality from which the electricity prices must be retrieved. Can be in form of city ID or city name.
        If not given, queries the ELCOM database for the prices in every municipality.
    category : str
        Category from which the electricity prices must be retrieved.
        If not given, prices are given for every consumer category.
    tva : float
        Scaling factor for the resulting prices, initialized as the normal TVA.
    export_path : str
        If given, export the prices with the parameter required at the path.

    Returns
    -------
    pd.DataFrame with the electricity price and its components.

    See also
    --------
    get_prices_from_elcom_by_canton : To retrieve the ELCOM prices by canton.
    get_injection_prices : To obtain the injection prices instead.

    Notes
    -----
    - A ``QBuildingsReader`` object can be passed to 'city' for automatic localization.
    - List and description of the available categories are available at the `ELCOM website <https://www.prix-electricite.elcom.admin.ch/>`_.
    - The TVA on electricity changed in 2024, from 7.7% to 8.1%.

    Examples
    --------
    >>> prices = electricity_prices.get_prices_from_elcom_by_city(city='Geneva', category='H4')
    >>> prices
        Year  City   Category  ...  community_fees  aidfee  finalcosts
    0  2024  Geneva       H4  ...         1.42824     2.3   30.925972
    [1 rows x 9 columns]
    """

    city_link = 'https://ld.admin.ch/municipality/'
    cat_link = 'https://energy.ld.admin.ch/elcom/electricityprice/category/'

    if isinstance(city, QBuildingsReader):
        try:
            city = city.data['transformers']['id_city'][0]
        except KeyError:
            city = 6621  # if id_city is missing from database, set Geneva by default
        city_query = '<' + city_link + str(city) + '>'
    elif isinstance(city, str):
        cities = pd.read_csv(os.path.join(path_to_elcom, 'correspondance_table_municipality_operator.csv'), index_col=0)
        mask = (cities['id_city'] == city) + (cities['commune'] == city)
        cities = cities[mask]
        if cities.empty:
            print('No corresponding city with that name')
        city = cities.iloc[0]['id_city']
        city_query = '<' + city_link + str(city) + '>'
    elif isinstance(city, int):
        city_query = '<' + city_link + str(city) + '>'
    else:
        city_query = "?id_city"

    valid_cat = ['H1', 'H2', 'H3', 'H4', 'H5', 'H6', 'H7', 'H8', 'C1', 'C2', 'C3', 'C4', 'C5', 'C6', 'C7']
    if category in valid_cat:
        cat_query = '<' + cat_link + category + '>'
    elif category is None:
        cat_query = "?Category"
    else:
        raise ValueError("The category asked is not a valid one from the elcom.\n"
                         f"Please pick among {valid_cat}.\n")

    query_template = """
    SELECT ?Year ?City ?Provider ?Category ?totalcosts ?energy ?grid ?community_fees ?aidfee
    WHERE {
        <https://energy.ld.admin.ch/elcom/electricityprice> cube:observationSet ?observationSet0 .
        ?observationSet0 cube:observation ?observation .

        ?observation
          elcom:category category_to_replace;
          elcom:municipality city_to_replace;
          elcom:operator ?operator;
          elcom:period "year_to_replace"^^<http://www.w3.org/2001/XMLSchema#gYear>;
          elcom:product <https://energy.ld.admin.ch/elcom/electricityprice/product/standard>;
          elcom:total ?totalcosts;
          elcom:gridusage ?grid;
          elcom:energy ?energy;
          elcom:charge ?community_fees;
          elcom:aidfee ?aidfee.
          
      OPTIONAL {
        city_to_replace <https://schema.org/name> ?commune_0 .
        FILTER (
          LANGMATCHES(LANG(?commune_0), "fr")
        )
      }
      OPTIONAL {
        city_to_replace <https://schema.org/name> ?commune_1 .
        FILTER (
          LANGMATCHES(LANG(?commune_1), "de")
        )
      }
      OPTIONAL {
        city_to_replace <https://schema.org/name> ?commune_2 .
        FILTER (
          LANGMATCHES(LANG(?commune_2), "it")
        )
      }
      OPTIONAL {
        city_to_replace <https://schema.org/name> ?commune_3 .
        FILTER (
          LANGMATCHES(LANG(?commune_3), "en")
        )
      }
      OPTIONAL {
        city_to_replace <https://schema.org/name> ?commune_4 .
        FILTER (
          (LANG(?commune_4) = "")
        )
      }
    BIND(COALESCE(?commune_0, ?commune_1, ?commune_2, ?commune_3, ?commune_4) AS ?City)
          
      OPTIONAL {
        ?operator <https://schema.org/name> ?operator_0 .
        FILTER (
          LANGMATCHES(LANG(?operator_0), "fr")
        )
      }
      OPTIONAL {
        ?operator <https://schema.org/name> ?operator_1 .
        FILTER (
          LANGMATCHES(LANG(?operator_1), "de")
        )
      }
      OPTIONAL {
        ?operator <https://schema.org/name> ?operator_2 .
        FILTER (
          LANGMATCHES(LANG(?operator_2), "it")
        )
      }
      OPTIONAL {
        ?operator <https://schema.org/name> ?operator_3 .
        FILTER (
          LANGMATCHES(LANG(?operator_3), "en")
        )
      }
      OPTIONAL {
        ?operator <https://schema.org/name> ?operator_4 .
        FILTER (
          (LANG(?operator_4) = "")
        )
      }
      BIND(COALESCE(?operator_0, ?operator_1, ?operator_2, ?operator_3, ?operator_4) AS ?Provider)
            
    }
    """

    query = query_template.replace('year_to_replace', str(year))
    query = query.replace('city_to_replace', city_query)
    query = query.replace('category_to_replace', str(cat_query))
    prices = sparql.send_query(query)
    if prices.loc[0, 'Category'] is None:
        prices['Category'] = category
    else:
        prices['Category'] = prices['Category'].apply(lambda row: row.split(cat_link)[1])
    prices['Year'] = year

    # Define custom sorting order
    custom_order = {'H1': 1, 'H2': 2, 'H3': 3, 'H4': 4, 'H5': 5, 'H6': 6, 'H7': 7, 'H8': 8,
                    'C1': 9, 'C2': 10, 'C3': 11, 'C4': 12, 'C5': 13, 'C6': 14, 'C7': 15}
    prices['Category'] = prices['Category'].map(custom_order)

    # Sort the DataFrame based on the "Category" column
    prices = prices.sort_values(by='Category').reset_index(drop=True)

    # Convert back to original category format
    prices['Category'] = prices['Category'].map({v: k for k, v in custom_order.items()})

    if tva is None:
        if year >= 2024:
            tva = 1.081
        else:
            tva = 1.077

    prices['finalcosts'] = tva * prices['totalcosts']

    if isinstance(export_path, str):
        prices.to_csv(os.path.realpath(export_path))

    return prices


def get_vese_key():
    try:
        response = rq.get("https://ipese-lectures.epfl.ch/static/reho.json", timeout=1)
        if response.status_code == 200:
            json_data = response.json()
            api_key_value = json_data.get("API_VESE_KEY")
            return api_key_value
        else:
            return f"Error: {response.status_code}"
    except Exception:
        return False


def get_injection_prices(city=None, year=2024, category=None, tva=None):
    """
    Retrieves injection prices from the `pvtarif.ch <https://www.vese.ch/fr/pvtarif/>`_  API.

    The year, municipality and consumer category can be given to query at a more precise level.
    TVA is applied by default or can be adapted as a scaling factor.

    Parameters
    ----------
    city : str or None, optional
        The city for which to retrieve injection prices. If None, prices for all cities will be retrieved.
    year : int, optional
        The year for which to retrieve injection prices. Default is 2024.
    category : str or None, optional
        The energy category for which to retrieve injection prices. If None, prices for the first power category are given.
    tva : float or None, optional
        The Value Added Tax (TVA) multiplier to apply to the total costs. If None, the default TVA value is used.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing injection prices information for each city.

    Raises
    ------
    ExecutionError
        Raised if there is an issue with the HTTP request to the PVTarif API.

    See also
    --------
    get_prices_from_elcom_by_city : To retrieve the ELCOM prices by city.
    get_prices_from_elcom_by_canton : To retrieve the ELCOM prices by canton.

    Notes
    -----
    - The data are not realibly available before 2017.
    - The category corresponds to the one from `ELCOM <https://www.prix-electricite.elcom.admin.ch/>`_.
    - The TVA on electricity changed in 2024, from 7.7% to 8.1%.

    Example
    -------
    >>> retribution_prices = get_injection_prices(year=2023, city='Basel')
    >>> retribution_prices.columns
    Index(['id_city', 'municipality', 'id_operator', 'operator', 'federal_tariff',
       'origin_bonus', 'totalcosts', 'finalcosts'],
      dtype='object')
    >>> retribution_prices
        id_city municipality  id_operator  ... origin_bonus  totalcosts  finalcosts
    1914     2701        Basel          624  ...          0.0        13.0        14.0
    [1 rows x 8 columns]
    """
    # Retrieve license key
    load_dotenv()
    if 'API_VESE_KEY' not in os.environ:
        license_key = get_vese_key()
        if not license_key:
            raise UserWarning("You need a key from VESE to access the injection prices")
    else:
        license_key = os.environ["API_VESE_KEY"]
    if len(str(year)) == 4:
        year = year % 100

    # Find the operator at that commune
    communes = get_providers_by_municipality_id(city)
    injection_prices = pd.DataFrame()
    for commune in communes.itertuples():
        url_grd = f'https://opendata.vese.ch/pvtarif/api/ClientService.php?mode=evu&evuId={commune.id_operator}' \
                  f'&year={year}&licenseKey={license_key}'
        response = rq.get(url_grd)
        if response.status_code == 200:
            json_data = response.content.decode('utf-8')
            json_data = json.loads(json_data)
            if not json_data['valid']:
                print(f'{json_data["code"]}: {json_data["details"]}')
                continue
        else:
            raise ExecutionError(f"{response.status_code}")

        try:
            # TODO: add a code to adapt the prices to the category given, as a function of what is defined
            commune_price = {'id_city': commune.id_city, 'municipality': commune.commune,
                             'id_operator': int(json_data['nrElcom']), 'operator': json_data['nomEw'],
                             'federal_tariff': float(json_data['energy1']), 'origin_bonus': float(json_data['eco1']),
                             'totalcosts': float(json_data['energy1']) + float(json_data['eco1'])
                             }
        except:
            continue
        commune_price = pd.DataFrame(commune_price, index=[commune.Index])
        injection_prices = pd.concat([injection_prices, commune_price])
    if tva is None:
        if year >= 2024:
            tva = 1.081
        else:
            tva = 1.077
    injection_prices['finalcosts'] = round(injection_prices['totalcosts'] * tva, 2)

    return injection_prices


def get_electricity_prices(city, year=2024, category=None, tva=None):
    """
    Builds a DataFrame with the electricity prices (demand and supply) ready to use for REHO.

    It calls `get_prices_from_elcom_by_city` and `get_injection_prices` and merges the two.

    Parameters
    ----------
    year : int
        Year from which the electricity prices must be retrieved.
    city : str/int
        Municipality from which the electricity prices must be retrieved. Can be in form of city ID or city name.
        If not given, queries the ELCOM database for the prices in every municipality.
    category : str
        Category from which the electricity prices must be retrieved.
        If not given, prices are given for every consumer category.
    tva : float
        Scaling factor for the resulting prices, initialized as the normal TVA.

    Returns
    -------
    pd.DataFrame with prices for the given parameters which columns are ['Year', 'City', 'Provider', 'Category', 'Elec_demand_cts_kWh', 'Elec_supply_cts_kWh'].

    See also
    --------
    get_prices_from_elcom_by_city : To retrieve the ELCOM prices by city.
    get_injection_prices : To obtain the injection prices instead.

    Examples
    --------
    >>> get_electricity_prices(year=2017, city='Genève')
        Year    City  ... Elec_demand_cts_kWh Elec_supply_cts_kWh
    0   2017  Genève  ...           22.216512               12.92
    1   2017  Genève  ...           21.887440               12.92
    2   2017  Genève  ...           19.197310               12.92
    3   2017  Genève  ...           21.596772               12.92
    4   2017  Genève  ...           19.367290               12.92
    5   2017  Genève  ...           16.895382               12.92
    6   2017  Genève  ...           19.316962               12.92
    7   2017  Genève  ...           21.598712               12.92
    8   2017  Genève  ...           23.155285               12.92
    9   2017  Genève  ...           23.548888               12.92
    10  2017  Genève  ...           21.866694               12.92
    11  2017  Genève  ...           20.781146               12.92
    12  2017  Genève  ...           22.345242               12.92
    13  2017  Genève  ...           17.155742               12.92
    14  2017  Genève  ...           15.900062               12.92
    [15 rows x 6 columns]

    """
    demand_prices = get_prices_from_elcom_by_city(year=year, city=city, category=category, tva=tva)
    supply_prices = get_injection_prices(city=city, year=year, category=category, tva=tva)

    electricity_prices = demand_prices[['Year', 'City', 'Provider', 'Category']].copy()
    electricity_prices['Elec_demand_cts_kWh'] = demand_prices['finalcosts']
    electricity_prices.loc[:, 'Elec_supply_cts_kWh'] = supply_prices['finalcosts'].iloc[0]

    return electricity_prices


if __name__ == '__main__':
    injection = get_injection_prices(city='Basel', year=2023)
    communes = get_electricity_prices(city='Geneva', year=2017, category=None)
    prices = get_prices_from_elcom_by_city(year=2023, city='Geneva', category=None, export_path='../../elec_prices_2024.csv')
    id_commune = 177
    get_injection_prices(692)
    # url_commune = f'https://opendata.vese.ch/pvtarif/api/ClientService.php?mode=muni&idofs={id_commune}&licenseKey={license_key}'



import re
import pandas as pd
import numpy as np
from reho.paths import *

__doc__ = """
Builds the dataframe for the visualization of annual flows from REHO results in the form of a Sankey diagram.
"""

# Colors and labels for units and layers
layout = pd.read_csv(os.path.join(path_to_plotting, 'layout.csv'), index_col='Name').dropna(how='all')


def update_label(source_name, target_name, df_label):
    """
        update labels of df_label if source_name or target_name not in index of df_label

        Parameters:
            source_name (string) : first name to update
            target_name (string) : second name to update
            df_label (df) : pd.DataFrame with labels

        Returns:
            df_label (df) updated
    """
    if not (source_name in df_label.index):  # create label 'source' if not existing yet
        df_label.loc[source_name, 'pos'] = len(df_label)
    if not (target_name in df_label.index):  # create label 'target' if not existing yet
        df_label.loc[target_name, 'pos'] = len(df_label)
    return df_label


def add_label_value(df_label, df_stv, precision, units):
    """
        add the values from df_stv to the labels of df_labels
        The value of the nodes are thus available in the nodes name for the sankey diagram

        Parameters:
            df_label (df) : pd.DataFrame of labels
            df_stv (df) : pd.DataFrame of source, target and value
            precision (int): precision of the displayed numbers (default = 2)
            units (string): unit of the values (default MWh)

        Returns:
            df label updated with the label values
    """
    df_source_value = pd.DataFrame()
    df_source_value.index = df_label.pos

    for i in list(df_source_value.index):
        source_val = df_stv.loc['value', df_stv.loc['source', :] == i].sum()
        target_val = df_stv.loc['value', df_stv.loc['target', :] == i].sum()
        # higher value on the node is displayed, (i.e. size of the box node on the sankey)
        df_source_value.loc[i, 'value'] = max(source_val, target_val)

    df_label = df_label.merge(df_source_value, left_on='pos', right_index=True)
    df_label.label = df_label.label + "\n" + df_label.value.round(precision).astype(str) + units
    return df_label


def add_flow(source, dest, layer, hub, dem_sup, df_annuals, df_label, df_stv, check_dest_2=False, dest_2=None, adjustment=0, fact=1):
    """
        Add an energy flow for the sankey diagramm for 'sankey_plot' according (a) cell(s) of df_annuals if cell not null

        Parameters:
            source (string) : name of the source
            dest (string) : name of the destination
            layer (string) : name of the layer of the considered cell(s)
            hub (string) : name of the hub of the considered cell(s)
            dem_sup (string) : 'Supply_MWh' or 'Demand_MWh', column to take (! no control)
            df_annuals (df) : df_annuals dataframe
            df_label (df) : df_label dataframe of labels
            df_stv (df) : df_stv dataframe of source,target,value
            check_dest_2 (bool) : if True dest_2 substitute dest (default false)
            dest_2 (string) : second possible destination (default none)
            adjustment (float) : offset added to the cell value (default 0)
            fact (float) : factor multiplied to the cell value (default 1)

        Returns:
            df_label updated (df), df_stv updated (df), value added (float, 0 if nothing added)
    """
    source_to_dest = df_annuals.loc[(df_annuals['Layer'] == layer) & (df_annuals['Hub'] == hub)][
        dem_sup].sum()  # .sum() to add all the values of the buildings if multiple buildings
    if source_to_dest != 0:  # if data available
        if check_dest_2:  # if True apply the second dest
            dest = dest_2
        df_label = update_label(source, dest, df_label)  # update label list
        df_stv[source + '_to_' + dest] = [df_label.loc[source, 'pos'],  # source, create a source to target column
                                          df_label.loc[dest, 'pos'],  # target
                                          float(fact * source_to_dest + adjustment)]  # value

    return df_label, df_stv, fact * source_to_dest + adjustment


def df_sankey(df_results, label='FR_long', color='ColorPastel', precision=2, units='MWh', display_label_value=True, scaling_factor=1):
    # Hypotheses :
    # 1. DHW demand taken as the supply of the watertank DHW
    # 2. no flow: electrical storage system to grid feed in, all to 'feed in electrical grid ' is from PV
    # 3. Small losses of NG, heat, wood,... between network and devices not accounted
    # 4. Electricity for 'Data heat' fully accounted as electricity consumption (Layer Data: not in sankey)
    # 5. Electricity produced by technologies can be stored (e.g. NG_cogen elec -> battery)

    # Multi building :
    # supported

    # !! MAKE SURE : all the possible technologies/sources/demands are in the list below (if not, risk that sth will be
    # not displayed, there is no check provided by this function for that)
    # List of supported technologies/sources/demands
    # Electrical storage device
    elec_storage_list = ['Battery', 'EV_district']
    # Manual handled devices (list below not used, just here for the information)
    # manual_device = ['PV', 'WaterTankSH']
    # Semi automatic handled devices
    semi_auto_device = ['HeatPump_Air', 'HeatPump_DHN', 'NG_Boiler', 'ThermalSolar', 'OIL_Boiler',
                        'ElectricalHeater_DHW', 'ElectricalHeater_SH', 'NG_Cogeneration', 'DHN_in',
                        'HeatPump_Lake', 'WOOD_Stove', 'HeatPump_Geothermal', 'Air_Conditioner',
                        'DataHeat_DHW']  # name must be the same as used by REHO
    # Network (electrical grid, oil network...) and end use demand (DHW, SH, elec appliances) handled automatically

    # Select only not null lines in df_annuals
    df_annuals = scaling_factor * df_results['df_Annuals']
    df_annuals = df_annuals.replace(0, np.nan)
    df_annuals = df_annuals.loc[df_annuals['Demand_MWh'].notnull() | df_annuals['Supply_MWh'].notnull()]
    df_annuals = df_annuals.replace(np.nan, 0).reset_index()

    # "Building" string management in data to deal with data uniformly regarding buildings
    for x in list(df_annuals.index):
        if df_annuals.loc[x, "Hub"].startswith('Building'):  # all Buildingx -> Building
            df_annuals.loc[x, "Hub"] = "Building"
        else:  # else remove '_Buildingx' of the hub name
            df_annuals.loc[x, "Hub"] = re.sub("_Building\d+", "", df_annuals.loc[x, "Hub"])

    # Creating the dataframe for the following
    df_label = pd.DataFrame(columns=['pos'])  # df for labels, position number (pos) and colors
    df_stv = pd.DataFrame(index=['source', 'target', 'value'])  # df for source,target,value

    # check if electricity storage
    elec_storage_use = False
    for elec_storage in elec_storage_list:
        if len(df_annuals.loc[(df_annuals['Layer'] == 'Electricity') & (df_annuals['Hub'] == elec_storage)]) != 0:
            elec_storage_use = True

    # check if watertank SH
    watertank_sh = False
    if len(df_annuals.loc[(df_annuals['Layer'] == 'SH') & (df_annuals['Hub'] == 'WaterTankSH')]) != 0:
        watertank_sh = True

    # 1 Elec Grid to Elec Consumption of the building(s) or before electricity storage (=Before Phase, bp) if present
    df_label, df_stv, _ = add_flow('Electrical_grid', 'Electrical_consumption', 'Electricity', 'Network', 'Supply_MWh',
                                   df_annuals, df_label, df_stv, elec_storage_use, 'Electrical_consumption_bp')

    # 2 Elec Cons to Elec Appliances
    df_label, df_stv, _ = add_flow('Electrical_consumption', 'Electrical_appliances', 'Electricity', 'Building',
                                   'Demand_MWh', df_annuals, df_label, df_stv)

    # 3 PV to Electrical Grid feed in
    df_label, df_stv, pv_to_egf = add_flow('PV', 'Electrical_grid_feed_in', 'Electricity', 'Network',
                                           'Demand_MWh', df_annuals, df_label, df_stv)

    # 4 PV to Electrical cons (before battery if present)
    df_label, df_stv, _ = add_flow('PV', 'Electrical_consumption', 'Electricity', 'PV',
                                   'Supply_MWh', df_annuals, df_label, df_stv, elec_storage_use, 'Electrical_consumption_bp',
                                   -pv_to_egf)

    # 5 WaterTankSH to SH
    df_label, df_stv, wtsh_to_sh = add_flow('WaterTankSH', 'SH', 'SH', 'WaterTankSH', 'Supply_MWh',
                                            df_annuals, df_label, df_stv)

    # Calculation of Losses in the watertank
    heat_tot_sh = df_annuals.loc[(df_annuals['Layer'] == 'SH') & (df_annuals['Hub'] == 'Building')].Demand_MWh.sum()
    heat_tot_sup = df_annuals.loc[(df_annuals['Layer'] == 'SH')].Supply_MWh.sum() - wtsh_to_sh
    heat_loss_wt = heat_tot_sup - heat_tot_sh

    # 6 SH Heat to Watertank SH if watertank
    df_label, df_stv, _ = add_flow('SH_heat', 'WaterTankSH', 'SH', 'WaterTankSH', 'Supply_MWh',
                                   df_annuals, df_label, df_stv, adjustment=heat_loss_wt)

    # 7 SH Heat to SH if watertank
    df_label, df_stv, _ = add_flow('SH_heat', 'SH', 'SH', 'WaterTankSH', 'Supply_MWh',
                                   df_annuals, df_label, df_stv, adjustment=heat_tot_sup - heat_loss_wt, fact=-1)

    elec_storage_energy_in = 0
    elec_storage_energy_out = 0
    for elec_storage in elec_storage_list:
        # 8 Electrical cons before elec storage to storage device
        df_label, df_stv, dev_flow_in = add_flow('Electrical_consumption_bp', elec_storage, 'Electricity', elec_storage,
                                                 'Demand_MWh', df_annuals, df_label, df_stv)
        elec_storage_energy_in += dev_flow_in

        # 9 storage device to Electrical cons
        df_label, df_stv, dev_flow_out = add_flow(elec_storage, 'Electrical_consumption', 'Electricity', elec_storage,
                                                  'Supply_MWh', df_annuals, df_label, df_stv)
        elec_storage_energy_out += dev_flow_out

    # 10 Electrical cons before elec storage to Electr cons
    if elec_storage_use:
        df_label = update_label('Electrical_consumption_bp', 'Electrical_consumption', df_label)
        Ec_after_bp = df_annuals.loc[(df_annuals['Layer'] == 'Electricity') & (df_annuals['Hub'] != 'Network')].Demand_MWh.sum()
        Ec_after_bp -= elec_storage_energy_out
        df_stv['Electrical_consumption_bp_to_Electrical_consumption'] = [df_label.loc['Electrical_consumption_bp', 'pos'],
                                                                         df_label.loc['Electrical_consumption', 'pos'],
                                                                         float(Ec_after_bp - elec_storage_energy_in)]

    # Semi-Auto for the followings devices
    for device in semi_auto_device:
        
        # 1 Ele Cons to Device
        df_label, df_stv, _ = add_flow('Electrical_consumption', device, 'Electricity', device, 'Demand_MWh',
                                       df_annuals, df_label, df_stv)

        # 2 Heat to Device
        df_label, df_stv, _ = add_flow('Heat', device, 'Heat', device, 'Demand_MWh',
                                       df_annuals, df_label, df_stv)

        # 3 NG to Device
        df_label, df_stv, _ = add_flow('NaturalGas', device, 'NaturalGas', device, 'Demand_MWh',
                                       df_annuals, df_label, df_stv)

        # 3 Wood to Device
        df_label, df_stv, _ = add_flow('Wood', device, 'Wood', device, 'Demand_MWh',
                                       df_annuals, df_label, df_stv)

        # 4 Oil to Device
        df_label, df_stv, _ = add_flow('Oil', device, 'Oil', device, 'Demand_MWh',
                                       df_annuals, df_label, df_stv)

        # 5 Device to DHW
        df_label, df_stv, _ = add_flow(device, 'DHW', 'DHW', device, 'Supply_MWh',
                                       df_annuals, df_label, df_stv)

        # 6 Device to SH Heat (if watertank) or to SH (if no watertank)
        df_label, df_stv, _ = add_flow(device, 'SH', 'SH', device, 'Supply_MWh',
                                       df_annuals, df_label, df_stv, watertank_sh, 'SH_heat')

        # 7 Device to Elec Consumption (or before electricity storage if present)
        df_label, df_stv, _ = add_flow(device, 'Electrical_consumption', 'Electricity', device, 'Supply_MWh',
                                       df_annuals, df_label, df_stv, elec_storage_use, 'Electrical_consumption_bp')

        # 8 Device to Cooling
        df_label, df_stv, _ = add_flow(device, 'Cooling', 'Cooling', device, 'Supply_MWh',
                                       df_annuals, df_label, df_stv)

    # df_label : add the label to display, the color and the label (node) values if selected
    df_label['label'] = layout[label]
    df_label['color'] = layout[color]
    if display_label_value:
        df_label = add_label_value(df_label, df_stv, precision, units)

    # data to export
    source = list(df_stv.loc['source'])
    target = list(df_stv.loc['target'])
    value = list(df_stv.loc['value'])
    value = [round(x, precision) for x in value]  # precision
    label = list(df_label.label)
    color = list(df_label.color)

    return source, target, value, label, color



import pandas as pd
import numpy as np
import scipy.spatial
from sklearn_extra.cluster import KMedoids
from sklearn.metrics import pairwise_distances


class ClusterClass:
    """
    This class executes the data reduction for meteorological data.
    It will execute a clustering for each number of clusters desired among an interval (nb_clusters), and select the optimal one according to the MAPE criterion (Maximum Average Pairwise Euclidean distance).

    Parameters
    ----------
    data : pd.DataFrame
        Annual weather data
    nb_clusters : list
        Interval for the number of clusters tested.

    """

    def __init__(self, data, nb_clusters=None, option=None, pd=None, outliers=None):
        super().__init__()
        if nb_clusters is None:
            self.nb_clusters = [8]
        else:
            self.nb_clusters = nb_clusters

        # org - original, nor - normalized, clu - clustered

        self.outlier_removal = outliers

        self.data_org = data  # input data- df: attribute vectors as columns
        self.data_nor = None
        self.attr_org = []  # matrix or attributes, arrays:  len(attr_org) = total number of periods per year, len(attr_org[0])= number of attributes x period duration
        self.attr_nor = []
        self.attr_clu = None
        self.modula = None
        self.mod_org = []  # modulo data, f.e. last day of year (24hrs) for  pd = 7*24
        self.mod_nor = []
        if option is None:
            self.option = {"year-to-day": True, "extreme": [(1, "min"), (2, "max")]}
        else:
            self.option = option

        self.results = {"idx": None}  # pd.DataFrame()

        self.kpis_clu = None
        self.nbr_opt = None

        if pd is None:  # period duration, f.e. 1 period = 1 day => pd =  24 (default)
            self.pd = 24
        else:
            self.pd = pd

    def run_clustering(self):
        self.__do_normalization()
        self.__execute_clustering()
        self.__compute_kpis()
        self.__select_optimal()

    def __do_normalization(self):
        # - Load data set
        self.data_nor = pd.DataFrame([], index=self.data_org.index, columns=self.data_org.columns)
        # - Create dissimilarity matrix
        for id, col in enumerate(self.data_org):
            self.data_nor.loc[:, col] = ((self.data_org.loc[:, col] - self.data_org.loc[:, col].min()) / (
                        self.data_org.loc[:, col].max() - self.data_org.loc[:, col].min()))  # normalize
            # - OPTION : re-arrange dataframe
            if self.option["year-to-day"]:
                nbr_p = int(self.data_nor.shape[0] // self.pd)  # ! integer division / number of periods default: 365
                self.modulo = int(self.data_nor.shape[0] % self.pd)
                stop = self.data_nor.shape[0] - self.modulo
                self.attr_org.append(np.reshape(self.data_org.loc[: (stop - 1), col].tolist(), (nbr_p, self.pd)))
                self.attr_nor.append(np.reshape(self.data_nor.loc[: (stop - 1), col].tolist(), (nbr_p, self.pd)))
                self.mod_org.append(self.data_org.loc[stop:, col].tolist())
                self.mod_nor.append(self.data_nor.loc[stop:, col].tolist())
            else:
                self.attr_org.append(np.reshape(self.data_org.loc[:, col].tolist(), (self.data_org.shape[0], 1)))
                self.attr_nor.append(np.reshape(self.data_nor.loc[:, col].tolist(), (self.data_org.shape[0], 1)))

        self.attr_org = np.hstack(self.attr_org)
        self.attr_nor = np.hstack(self.attr_nor)

    @staticmethod
    def __run_KMedoids(matrix, n_clusters):

        if len(matrix) == 1:  # check for trivial solution
            df_res = pd.DataFrame([1], columns=['1'])

        else:

            print('Applying algorithm for', n_clusters, 'clusters')

            dist_a = pairwise_distances(matrix, metric='sqeuclidean')
            kmedoids = KMedoids(n_clusters=n_clusters, method='pam', random_state=42)
            kmedoids.fit(dist_a)

            cluster_assignments = kmedoids.labels_
            medoid_indices = kmedoids.medoid_indices_

            label_to_medoid = dict(zip(np.unique(cluster_assignments), medoid_indices))
            mapped_cluster_assignments = np.vectorize(label_to_medoid.get)(cluster_assignments)

            df_res = pd.DataFrame()
            df_res[str(n_clusters)] = mapped_cluster_assignments

        return df_res

    def __execute_clustering(self):

        df_res = pd.DataFrame()  # initialize df for results

        for i, c in enumerate(self.nb_clusters):
            df = self.__run_KMedoids(self.attr_nor, c)
            df_res[str(c)] = df[str(c)]

        if self.outlier_removal:
            for i, c in enumerate(self.nb_clusters):
                ncluster = c

                medoids = df_res[str(ncluster)].unique()  # Warning periods start with 1, python with 0

                df_distance = pd.DataFrame()

                for med in medoids:
                    group = df_res[df_res[str(ncluster)] == med]
                    df_attr = pd.DataFrame(self.attr_nor)

                    group_matrix = df_attr.loc[group.index]  # all periods allocated to medoid
                    group_vector = group_matrix.loc[med - 1]  # get medoid as vector

                    np_group_matrix = group_matrix.to_numpy()
                    np_group_vector = group_vector.to_numpy().reshape(1, -1)

                    dist = scipy.spatial.distance.cdist(np_group_vector, np_group_matrix)  # calculate distance to medoid

                    df_med = pd.DataFrame(dist[0], index=group_matrix.index)

                    df_distance = df_distance.append(df_med)

                if 'nlargest' in self.outlier_removal.keys():
                    outliers = df_distance.nlargest(self.outlier_removal['nlargest'], 0)
                elif 'distance' in self.outlier_removal.keys():
                    outliers = df_distance[df_distance[0] > self.outlier_removal['distance']]

                else:
                    raise "outlier removal not specified use for example {'nlargest': 3} or {'distance':1}"

                without_outliers = np.delete(self.attr_nor, outliers.index, axis=0)
                df = self.__run_KMedoids(without_outliers, ncluster)

                s = df.apply(lambda x: self.__return_typical_periods_outliers(x, x[str(ncluster)], outliers.index), axis=1)
                s = s.set_index(0)
                for o in outliers.index:
                    s.at[o, 1] = o
                df_res[str(c)] = s[1].sort_index()

        df_res.columns.name = "iteration"
        for col in df_res.columns:
            # - Rename extreme day index
            df_res.loc[df_res.loc[:, col] == 0, col] = df_res.loc[df_res.loc[:, col] == 0, col].index
        self.results["idx"] = df_res

    @staticmethod
    def __return_typical_periods_outliers(idx, clusters, outliers):
        index = idx.name
        outliers = outliers.sort_values()
        outlier_position = []  # find outlier position in new index
        for c, id in enumerate(outliers):
            outlier_position.append(id - c)
        outlier_position = np.array(outlier_position)

        outliers_before_period = outlier_position[outlier_position <= clusters]
        outliers_before_index = outlier_position[outlier_position <= index]

        if len(outliers_before_period) == 0:
            cluster = clusters
            if len(outliers_before_index) == 0:
                period = index
            else:
                period = index + len(outliers_before_index)
        else:
            cluster = clusters + len(outliers_before_period)
            if len(outliers_before_index) == 0:
                period = index
            else:
                period = index + len(outliers_before_index)

        return pd.Series([period, cluster])

    def __compute_kpis(self):
        frm_kpis = []
        frm_clus = []
        # - Iterate through solutions
        for sol in self.results["idx"]:
            # - Assess extreme periods
            self.__compute_extreme(sol)
            # - Assess clustered year
            attr_clu, data_clu = self.__do_attr_clu(sol)
            # data_clu annual dataframe, build with clustered periods to reach 8760 rows
            # attr_clu matrix, same size as attr_org but with clustered attributes
            # - Assess KPIs
            pi = pd.DataFrame(index=pd.Index(["LDC", "MAE", "RMSD", "MAPE"], name="kpi"), columns=pd.Index(self.data_org.columns, name="dimension"))
            # - LDC error
            pi.loc["LDC", :] = sum(abs(np.sort(data_clu.values, axis=0) - np.sort(self.data_nor.values, axis=0))) / sum(self.data_nor.values)
            # - MAE
            pi.loc["MAE", :] = (1 / data_clu.shape[0]) * abs(self.data_nor - data_clu).sum()
            # - RMSD
            pi.loc["RMSD", :] = ((1 / data_clu.shape[0]) * ((self.data_nor - data_clu) ** 2).sum()).apply(np.sqrt)
            # - MASE
            # diff = abs(self.data_N.shift() - self.data_N).sum() * (df_N.shape[0]) / (df_N.shape[0] -1)
            pi.loc["MAPE", :] = (1 / data_clu.shape[0]) * abs(self.data_nor - data_clu).sum() / (self.data_nor.mean())
            # - Append
            frm_kpis.append(pi)
            frm_clus.append(data_clu)
        # - Assign
        self.kpis_clu = pd.concat(frm_kpis, keys=self.results["idx"].columns, names=["iteration"], axis=1)
        self.attr_clu = pd.concat(frm_clus, keys=self.results["idx"].columns, names=["iteration"], axis=1)

    def __do_attr_clu(self, sol):
        # - Create clustered data
        idx = [x - 1 for x in self.results["idx"].loc[:, sol].tolist()]  # - Warning python index starts with 0!
        attr_clu = []
        for id in idx:
            id = int(id)
            attr_clu.append(self.attr_nor[id, :])

        # attr_clu.append(np.hstack(self.mod_nor))
        attr_clu = np.array(attr_clu)

        # - OPTION : re-arrange dataframe
        if self.option["year-to-day"]:
            df = pd.DataFrame(columns=self.data_org.columns)
            # nbr_t = int(self.data_nor.shape[0] / 365)
            for num, id_col in enumerate(df):
                df[id_col] = np.reshape(attr_clu[:, int(self.pd * num):int(self.pd * (num + 1))], (self.data_org.shape[0] - self.modulo, 1)).flatten()
        else:
            df = pd.DataFrame(data=attr_clu, index=self.data_org.index, columns=self.data_org.columns)

        # create df from the modulo data and append to cluster
        df_mod = pd.DataFrame.from_dict(dict(zip(self.data_org.columns, self.mod_nor)))
        df = df.append(df_mod, ignore_index=True)

        return attr_clu, df

    def __select_optimal(self):
        # - Define local variables
        kpis = self.kpis_clu.stack(["dimension"])
        diff = pd.DataFrame(index=kpis.index, columns=kpis.columns)
        self.nbr_opt = max(self.nb_clusters)
        # - Iterate over solutions
        for id, n_k in enumerate(kpis):
            if id < kpis.shape[1] - 1:
                diff.iloc[:, id] = (kpis.iloc[:, id + 1] - kpis.iloc[:, id])
                condition = (diff.loc[("MAPE", slice(None)), n_k] > -0.01).all()  # optimum condition: MAPE doesn't improve more than 1%
                if condition:
                    self.nbr_opt = n_k
                    break
        # print(diff)
        print("N_k optimal: " + str(self.nbr_opt))

    def __compute_extreme(self, sol):

        def find_nearests(array, value):
            array = np.asarray(array)
            dif = array - value
            if array.ndim > 1:
                dif[dif < 0] = 1e3
                idx = np.where(dif == np.amin(dif))
                id = idx[0][0]
                val = array[id, :]
            else:
                idx = (np.abs(array - value)).argmin()
                id = idx
                val = array[idx]
            return val, id

        # - Iterate over extreme options
        for id_dim in self.option["extreme"]:
            # - OPTION : re-arrange dataframe
            if self.option["year-to-day"]:
                nbr_t = self.pd  # int(self.data_nor.shape[0] / 365)
                id_str = int((id_dim[0] - 1) * nbr_t)
                id_end = int(id_dim[0] * nbr_t)
            else:
                id_str = int(id_dim[0])
                id_end = int(id_dim[0])
            # - Alter index
            if id_dim[1] == "max":
                _, max_id = find_nearests(self.attr_org[:, id_str:id_end], np.percentile(self.attr_org[:, id_str:id_end], 99))
                self.results["idx"].loc[max_id, sol] = max_id + 1  # - Warning -> index starts from 1!
            else:
                _, min_id = find_nearests(self.attr_org[:, id_str:id_end], np.percentile(self.attr_org[:, id_str:id_end], 1))
                self.results["idx"].loc[min_id, sol] = min_id + 1  # - Warning -> index starts from 1!



from reho.paths import *
import reho.model.preprocessing.weather as weather
import pandas as pd
import numpy as np

__doc__ = """
Generates electric vehicle (EV) demand profiles.
"""


def generate_EV_plugged_out_profiles_district(cluster, df_Timestamp):
    """
    Computes hourly electric vehicle profiles for each typical day considering weekdays and weekends.
    The EV demand profiles are used to optimize EV electricity demand profiles with the `evehicle.mod` ampl model.

    Parameters
    ----------
    cluster : dict
        Define location district, number of periods, and number of timesteps.
    df_Timestamp : pd.DataFrame
        Information for clustering results, used to know the periods and period duration.

    Returns
    -------
    EV_plugged_out : array
        Hourly profile of the share of vehicles being plugged out of the district LV grid.
    EV_plugging_in : array
        Hourly profile of the share of vehicles connecting to the district LV grid.

    Notes
    ------
    Data are taken from `UK Department for Transport 2013
    <https://www.gov.uk/government/collections/energy-and-environment-statistics#publications>`_ and `Swiss Federal Statistical Office 2015
    <https://www.bfs.admin.ch/asset/fr/1840478>`_.

    """
    # TODO IMPLEMENTATION of flexible period duration
    File_ID = weather.get_cluster_file_ID(cluster)

    if 'W' in File_ID.split('_'):
        use_weekdays = True
    else:
        use_weekdays = False

    if use_weekdays:
        timestamp = np.loadtxt(os.path.join(path_to_clustering, 'timestamp_' + File_ID + '.dat'), usecols=(1, 2, 3),
                               skiprows=1)
        timestamp = pd.DataFrame(timestamp, columns=("Day", "Frequency", "Weekday"))
    else:
        df = df_Timestamp
        timestamp = df.fillna(1)  # only weekdays

    # Federal Office of Statistic, Comportement de la population en matiere de transports, 2015
    # profiles_weekday_CH = [0.01, 0.0, 0.0, 0.0, 0.0, 0.03, 0.1, 0.12, 0.11, 0.12, 0.12, 0.14, 0.13, 0.14, 0.13, 0.13, 0.18,
    #                        0.2, 0.15, 0.1, 0.07, 0.06, 0.05, 0.03]
    # profiles_weekend_CH = [0.01, 0.0, 0.0, 0.0, 0.0, 0.00, 0.0, 0.00, 0.03, 0.10, 0.12, 0.11, 0.12, 0.12, 0.14, 0.13, 0.14,
    #                        0.13, 0.15, 0.1, 0.07, 0.06, 0.05, 0.03]

    # UK department for transport Electric Chargepoint Analysis 2017: Domestics
    profiles_weekday = [0.0, 0.0, 0.0, 0.0, 0.01, 0.02, 0.08, 0.2, 0.31, 0.35, 0.38, 0.39, 0.39, 0.38, 0.38, 0.36, 0.31,
                        0.24, 0.18, 0.13, 0.09, 0.05, 0.03, 0.01]
    profiles_weekend = [0.0, 0.0, 0.0, 0.0, 0.01, 0.02, 0.08, 0.2, 0.31, 0.35, 0.38, 0.39, 0.39, 0.38, 0.38, 0.36, 0.31,
                        0.24, 0.18, 0.13, 0.09, 0.05, 0.03, 0.01]

    plugging_in_weekday = [0.017, 0.006, 0.004, 0.003, 0.004, 0.005, 0.01, 0.019, 0.03, 0.034, 0.036, 0.042, 0.043,
                           0.045, 0.047, 0.067, 0.088, 0.111, 0.109, 0.087, 0.069, 0.058, 0.043, 0.023]
    plugging_in_weekend = [0.017, 0.006, 0.004, 0.003, 0.004, 0.005, 0.01, 0.019, 0.03, 0.034, 0.036, 0.042, 0.043,
                           0.045, 0.047, 0.067, 0.088, 0.111, 0.109, 0.087, 0.069, 0.058, 0.043, 0.023]

    EV_plugged_out = []  # all vehicules not connected
    EV_plugging_in = []  # vehicules connecting at time t
    # iter over the typical periods
    for j, day in enumerate(list(timestamp.Weekday)[:-2]):
        if day == 0:
            profile = profiles_weekend
            profile_plug_in = plugging_in_weekend
        elif day == 1:
            profile = profiles_weekday
            profile_plug_in = plugging_in_weekday
            # profile = np.tile(profiles_weekday, 365).tolist() # workaround -  whole year profile
        else:
            raise "day type not possible"

        EV_plugged_out = EV_plugged_out + profile
        EV_plugging_in = EV_plugging_in + profile_plug_in

    EV_plugged_out = EV_plugged_out + [0.0, 0.0]  # extreme hours
    EV_plugged_out = np.array(EV_plugged_out)

    EV_plugging_in = EV_plugging_in + [0.0, 0.0]  # extreme hours
    EV_plugging_in = np.array(EV_plugging_in)

    return EV_plugged_out, EV_plugging_in



import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import interp1d
import math
import scipy as sp

__doc__ = """
Reconstructs data from a Pareto optimization to produce a rainbow plot.
"""


def linerize_equal_steps(x, y):
    x.name = 'x'
    y.name = 'y'
    step = (x.max() - x.min()) / len(x)
    df = pd.concat([x, y], names=['x', 'y'], axis=1)

    df_new = pd.DataFrame(columns=['y'])
    for s in np.arange(1, len(x) + 1):
        x_new = s * step
        ub = df[df.x >= x_new].x.min()
        lb = df[df.x < x_new].x.max()
        if math.isnan(ub):  # happens in case x_new == to last x in array
            y_new = df[df.x == lb].iloc[0].y  # just return last point
        else:
            upper_bound = df[df.x == ub].iloc[0]
            lower_bound = df[df.x == lb].iloc[0]

            m = ((upper_bound.y - lower_bound.y) / (upper_bound.x - lower_bound.x))
            b = lower_bound.y - m * lower_bound.x

            y_new = m * x_new + b
        df_new.at[x_new, 'y'] = y_new
    return df_new


def split_orientation(az):
    if (az < 45) or (az > 315):
        orientation = 'North'
    elif (az >= 45) and (az < 135):
        orientation = 'East'
    elif (az >= 135) and (az < 225):
        orientation = 'South'
    elif (az >= 225) and (az <= 315):
        orientation = 'West'
    else:
        orientation = 'Other'

    return orientation


def get_roof_max(df_nbr, hs_A):
    df_n = df_nbr.groupby(['Pareto_ID', 'Surface', 'Azimuth', 'Tilt']).sum()
    roofs = df_n[df_n.index.get_level_values(level='Tilt') != 90]

    # get total m2 PV
    total_roofs = roofs.groupby('Pareto_ID').sum()
    total_roofs['PV_m2'] = total_roofs.PVA_module_nbr * 1.6
    # get occupation based on azimuth orientation
    roofs = roofs.reset_index(level=['Azimuth'])
    roofs['Orientation'] = roofs.apply(lambda x: split_orientation(x['Azimuth']), axis=1)
    roofs = roofs.set_index('Orientation', append='True')
    roofs = roofs.groupby(['Pareto_ID', 'Orientation']).sum()
    max = roofs.PVA_module_nbr.groupby(level="Orientation").max()
    roofs['occupancy'] = roofs.PVA_module_nbr / max

    roofs_max = total_roofs.PV_m2.max()
    roofs_max_m2 = roofs_max / hs_A

    facades = df_n[df_n.index.get_level_values(level='Tilt') == 90]
    if not facades.empty:
        # get total m2 facades
        total_facades = facades.groupby('Pareto').sum()
        total_facades['PV_m2'] = total_facades.PVA_module_nbr * 1.6
        # get occupation based on azimuth orientation
        facades = facades.reset_index(level=['Azimuth'])
        facades['Orientation'] = facades.apply(lambda x: split_orientation(x['Azimuth']), axis=1)
        facades = facades.set_index('Orientation', append='True')
        facades = facades.groupby(['Pareto', 'Orientation']).sum()
        max = facades.PVA_module_nbr.max(level=1)
        facades['occupancy'] = facades.PVA_module_nbr / max
        total_surface = total_facades['PV_m2'] + total_roofs['PV_m2']
    else:
        total_surface = total_roofs['PV_m2']
    total_surface_m2 = total_surface / hs_A

    return total_surface_m2, roofs_max_m2


def plot_economical_feedin_price(bounds, resolution, Pareto):
    df_unit = return_district_result_object_dataframe(Pareto, 'df_Unit')
    df_KPIs = return_district_result_object_dataframe(Pareto, 'df_KPIs')
    hs_A = return_district_result_object_dataframe(Pareto, 'df_Buildings')["ERA"].xs(1, level=1).sum()

    df_nbr = return_district_result_object_dataframe(Pareto, 'df_PV_orientation')  # df_PVA_module_nbr, ["df_PV_orientation"]
    total_surface_m2, roof_max_m2 = get_roof_max(df_nbr, hs_A)
    # some data processing
    MWh_PV = {}
    for i in Pareto[0]:
        if type(Pareto[0][i]) == dict:
            id = ["PV" in id for id in Pareto[0][i]["df_Annuals"].xs("Electricity")["Supply_MWh"].index]
            MWh_PV[i] = Pareto[0][i]["df_Annuals"].xs("Electricity")["Supply_MWh"][id].sum()
        else:
            id = ["PV" in id for id in Pareto[0][i].df_Annuals.xs("Electricity")["Supply_MWh"].index]
            MWh_PV[i] = Pareto[0][i].df_Annuals.xs("Electricity")["Supply_MWh"][id].sum()
    MWh_PV = pd.DataFrame.from_dict(MWh_PV, orient="Index")[0]
    total_surface_m2.index.name = "Pareto"

    df_K = df_KPIs.xs((0, 'Network'), level=('Scn_ID', 'Hub'))
    SC = df_K.SC
    MWh_PV.index = SC.index

    MWh_SC = MWh_PV.mul(SC)
    MWh_SC = MWh_SC.fillna(0)
    MWh_exp = MWh_PV - MWh_SC
    MWh_exp = MWh_exp.fillna(0)

    PV_kWyr = MWh_PV
    PV = df_unit[df_unit.index.get_level_values('Unit').str.contains('PV')]
    PV = PV.groupby('Pareto_ID').sum()
    BAT = df_unit[df_unit.index.get_level_values('Unit').str.contains('Battery')]
    BAT = BAT.groupby('Pareto_ID').sum()
    PV_CHF = PV.Costs_Unit_inv + BAT.Costs_Unit_inv
    PV_CHF[PV_CHF < 0] = np.float64(0)

    PV_CHF_kWyr = PV_CHF.div(PV_kWyr)
    PV_CHF_kWyr = PV_CHF_kWyr.fillna(0)

    # linearize PV design trends (PV capacity, MWh exported and imported)
    PV_CHF_kWyr = linerize_equal_steps(total_surface_m2, PV_CHF_kWyr).y
    MWh_SC = linerize_equal_steps(total_surface_m2, MWh_SC).y
    MWh_exp = linerize_equal_steps(total_surface_m2, MWh_exp).y

    # find PV_kW_yr where all roofs are full
    PV_kWyr_function = linerize_equal_steps(total_surface_m2, PV_kWyr)
    PV_kWyr = PV_kWyr_function.y

    No_feed_in = resolution
    No_retail = resolution
    min_feed = bounds["feed_in"][0]
    max_feed = bounds["feed_in"][1]
    min_demand = bounds["retail"][0]
    max_demand = bounds["retail"][1]
    feed_in_prices = np.linspace(min_feed, max_feed, No_feed_in)
    retail_prices = np.linspace(min_demand, max_demand, No_retail)

    # build matrix economic revenues for each combination feed-in / retail tariffs
    zero_data = np.zeros(shape=(No_feed_in, No_retail))
    df_plot_last = pd.DataFrame(zero_data, columns=retail_prices)
    df_plot_inv_induced = pd.DataFrame(zero_data, columns=retail_prices)
    df_plot_first = pd.DataFrame(zero_data, columns=retail_prices)
    for d in retail_prices:
        y_last = np.array([])
        y_inv_induced = np.array([])
        y_first = np.array([])
        for x in feed_in_prices:
            AR_x = (MWh_exp * x * 1000 + MWh_SC * d * 1000) / PV_kWyr
            Revenues = AR_x - PV_CHF_kWyr
            # last point which is economic = last point which has "positive value" so we need to find the 0 point. revenues with discrete cost can have 2 - we need to find the second one!
            max_index = Revenues[Revenues == Revenues.max()].index[0]

            # MAX POINT WHERE it corosses the last time
            if (Revenues.loc[max_index:] < 0).any():  # there can be max profit - but it is always economic->check if revenue gets negative
                if (Revenues.loc[max_index:] < 0).all():  # it is never "positiv" always uneconomic
                    last_economic_point = np.nan
                    inv_induced = np.nan
                else:
                    f = interp1d(Revenues.loc[max_index:], PV_kWyr.loc[max_index:])
                    last_economic_point = f(0)
                    f2 = interp1d(PV_kWyr.loc[max_index:], PV_CHF_kWyr.loc[max_index:])
                    inv_induced = f2(last_economic_point.mean())
            else:
                last_economic_point = -1.0
                inv_induced = -1.0
            # MIN POINT WHERE it corosses the first time
            if (Revenues.loc[:max_index] < 0).any():
                if (Revenues.loc[:max_index] < 0).all():
                    first_economic_point = np.nan
                else:
                    f = interp1d(Revenues.loc[:max_index], PV_kWyr.loc[:max_index])
                    first_economic_point = f(0)
            else:
                first_economic_point = np.nan

            y_last = np.append(y_last, last_economic_point)
            y_inv_induced = np.append(y_inv_induced, inv_induced)
            y_first = np.append(y_first, first_economic_point)
        df_plot_last[d] = y_last
        df_plot_inv_induced[d] = y_inv_induced
        df_plot_first[d] = y_first

    # df_plot_first = df_plot_first.set_index(feed_in_prices)
    df_plot_last = df_plot_last.set_index(feed_in_prices)
    df_plot_inv_induced = df_plot_inv_induced.set_index(feed_in_prices)

    return df_plot_last, feed_in_prices, No_feed_in, df_plot_inv_induced


def plot_rainbow(df_plot_last, feed_in_prices, No_feed_in, df_plot_inv_induced, bounds, save_fig, plot_type):

    fig, ax = plt.subplots()
    for demand_price in df_plot_last.columns:
        if plot_type == "demand_feed_in_E_gen_pv":
            if all(pd.isnull(df_plot_last[demand_price].values)):
                c_list = ["white"] * No_feed_in
            else:
                c_list = df_plot_last[demand_price].values
            cs = ax.scatter(feed_in_prices, np.repeat(demand_price, No_feed_in), c=c_list, s=5, cmap=cm, vmin=df_plot_last.min().min(),
                            vmax=df_plot_last.max().max())
        elif plot_type == "E_gen_pv_demand_feed_in":
            cs = ax.scatter(np.repeat(demand_price, No_feed_in), df_plot_last[demand_price].values, c=feed_in_prices, s=5, cmap=cm, vmin=bounds["feed_in"][0],
                            vmax=bounds["feed_in"][1])
        elif plot_type == "invest_demand_feed_in":
            cs = ax.scatter(np.repeat(demand_price, No_feed_in), df_plot_inv_induced[demand_price].values, c=feed_in_prices, s=5, cmap=cm,
                            vmin=bounds["feed_in"][0], vmax=bounds["feed_in"][1])

    ax.annotate(' all PV investments \n economic feasible', (0.08, 0.22), zorder=10)
    cbar = fig.colorbar(cs)

    if plot_type == "demand_feed_in_E_gen_pv":
        ax.set_xlim(bounds["feed_in"][0], bounds["feed_in"][1])
        ax.set_ylim(bounds["retail"][0], bounds["retail"][1])
        ax.set_xlabel('feed-in tariff  [CHF/kWh]', fontsize=14)
        ax.set_ylabel('retail tariff [CHF/kWh]', fontsize=14)
        cbar.ax.set_ylabel('PV electricity generated [TWh/yr]', fontsize=14)

    elif plot_type == "E_gen_pv_demand_feed_in":
        ax.set_xlim(bounds["retail"][0], bounds["retail"][1])
        ax.set_ylim(df_plot_last.min().min(), df_plot_last.max().max())
        ax.set_xlabel('retail tariff [CHF/kWh]')
        ax.set_ylabel('PV electricity generated [TWh/yr]')
        cbar.ax.set_ylabel('feed-in tariff  [CHF/kWh]')

    elif plot_type == "invest_demand_feed_in":
        ax.set_xlim(bounds["retail"][0], bounds["retail"][1])
        ax.set_ylim(df_plot_inv_induced.min().min(), df_plot_inv_induced.max().max())
        ax.set_xlabel('retail tariff [CHF/kWh]')
        ax.set_ylabel('investment [CHF/yr]')
        cbar.ax.set_ylabel('feed-in tariff  [CHF/kWh]')

    # ax.legend(loc='upper center', bbox_to_anchor=(0.5, -0.2), frameon=False, ncol=2)

    if save_fig:
        plt.tight_layout()
        format = 'png'
        plt.savefig(('PV_rainbow_CH' + '.' + format), format=format, dpi=300)
    plt.show()
    return


def return_district_result_object_dataframe(dict_results, result_dataframe):
    if result_dataframe in dir(dict_results[0][1]):
        t = {(j, k): getattr(dict_results[j][k], result_dataframe)
             for j in dict_results.keys()
             for k in dict_results[j].keys()}
    else:
        t = {(j, k): dict_results[j][k][result_dataframe]
             for j in dict_results.keys()
             for k in dict_results[j].keys()}

    df_district_results = pd.concat(t.values(), keys=t.keys(), names=['Scn_ID', 'Pareto_ID'], axis=0)
    df_district_results = df_district_results.sort_index()

    return df_district_results


def plot_rainbow_CH(Pareto_list, surface_district, surface_CH, CH_roof_area=None, bounds=None, resolution=100, std_filter=1.5,
                    plot_type="demand_feed_in_E_gen_pv", filter=0):
    if bounds is None:
        bounds = {"feed_in": [0.0, 0.15], "retail": [0.0, 0.30]}

    df_plot_last = {}
    df_plot_inv_induced = {}

    # get data
    for n, tr_id in enumerate(Pareto_list):

        print("district:", n, " processed")
        df_plot_last[n], feed_in_prices, No_feed_in, df_plot_inv_induced[n] = plot_economical_feedin_price(bounds, resolution, Pareto_list)

        # get last economic retail tariffs
        last_id = []
        for idx, col in enumerate(df_plot_last[n]):
            if any(df_plot_last[n][col].notnull()):
                df_not_nan = df_plot_last[n][col][df_plot_last[n][col].notnull()]
                if any(df_plot_last[n][col] > 0):
                    df_not_nan = df_not_nan[df_not_nan > 0]
                    last_id = np.concatenate([last_id, [df_not_nan.index[-1]]])
                else:
                    last_id = np.concatenate([last_id, [df_not_nan.index[0]]])
            else:
                last_id = np.concatenate([last_id, [np.nan]])
        last_id = pd.DataFrame(last_id)
        last_id[0] = last_id[0].interpolate()

        for idx, col in enumerate(df_plot_last[n]):
            df_plot_last[n][col] = df_plot_last[n][col].replace(np.nan, 0)
            index_max_PV = df_plot_last[n][col].loc[last_id.loc[idx][0]:].index[1:]
            df_plot_last[n][col].loc[index_max_PV] = df_plot_last[n].max().max()

    # aggregate data and extrapolate to CH
    df_plot_last_CH = df_plot_last[0] * 0
    for n, tr_id in enumerate(Pareto_list):
        if CH_roof_area is not None:
            ratio = list(surface_CH.values()) / np.sum(list(surface_CH.values()))
            df_plot_last_CH = df_plot_last_CH + df_plot_last[n] / surface_district[tr_id] * ratio[n] * CH_roof_area
        else:
            df_plot_last_CH = df_plot_last_CH + df_plot_last[n] / surface_district[tr_id] * surface_CH[tr_id]

    for col in df_plot_last_CH:
        df_plot_last_CH[col][df_plot_last_CH[col] < filter] = 0.0
        df_plot_last_CH = df_plot_last_CH.astype({col: "float"})

    df_plot_last_CH = sp.ndimage.filters.gaussian_filter(df_plot_last_CH, std_filter, mode='nearest')
    df_plot_last_CH = pd.DataFrame(df_plot_last_CH)
    df_plot_last_CH.columns = np.linspace(bounds["retail"][0], bounds["retail"][1], No_feed_in)
    df_plot_last_CH.index = np.linspace(bounds["feed_in"][0], bounds["feed_in"][1], No_feed_in)

    # create bounds
    max_value = df_plot_last_CH.max().max()

    for col in df_plot_last_CH:
        df_plot_last_CH[col].mask(df_plot_last_CH[col] < 0.03, np.nan, inplace=True)
        df_plot_last_CH[col].mask(df_plot_last_CH[col] > max_value - 0.1, np.nan, inplace=True)

    plot_rainbow(df_plot_last_CH, feed_in_prices, No_feed_in, df_plot_inv_induced, bounds, save_fig=True, plot_type=plot_type)


if __name__ == '__main__':
    plt.rcParams.update({'font.size': 14})
    cm = plt.cm.get_cmap('Spectral')

    Pareto = pd.read_pickle("results.pickle")

    surfaces_districts = {0: 100, 1: 200, 2: 300, 3: 400, 4: 500}  # surface of the typical districts in m2
    surfaces_country = {0: 1, 1: 2, 2: 3, 3: 4, 4: 5}  # representative surface of the typical districts at the national scale in 10e6 m2
    data = {i: Pareto[n] for i, n in enumerate(Pareto)}

    plot_rainbow_CH(data, surface_district=surfaces_districts, surface_CH=surfaces_country,
                    resolution=140, std_filter=1.5, plot_type="demand_feed_in_E_gen_pv", filter=6)



import pandas as pd
import numpy as np

__doc__ = """
Collects data from the `SIA Swiss norms <https://www.sia.ch/fr/services/sia-norm/>`_ , which are used to distinguish between eight different building types in their usage and behavior.
"""


def read_sia2024_rooms_sia380_1(digit, df_SIA_380):
    dict_affiliation2digit = {'collective housing': 'I',
                              'individual housing': 'II',
                              'administrative': 'III',
                              'school': 'IV',
                              'commercial': 'V',
                              'restaurant': 'VI',
                              'gathering places': 'VII',
                              'hospital': 'VIII',
                              'industry': 'IX',
                              'shed, warehouse': 'X',
                              'sports facilities': 'XI',
                              'covered swimming-pool': 'XII',
                              'other': 'XIII'}

    df_SIA_380 = df_SIA_380.rename(columns=dict_affiliation2digit)

    return df_SIA_380[digit]


def read_sia_2024_profiles(status, df):
    df_el_appliance = df["calculs"].iloc[:, 1:25]
    df_el_ap_norm = df_el_appliance.div(df_el_appliance.max(axis=1), axis=0)  # normalize profile
    df_el_ap_norm = df_el_ap_norm.fillna(0)

    # status allows to scale the domestic electricity consumption
    if status == 'existing':  # high electricity consumption
        df_existing = df["data"].iloc[:, 38]
        df_el_appliance = df_el_ap_norm.mul(df_existing.values, axis=0)
    elif status == 'standard':  # medium electricity consumption
        df_standard = df["data"].iloc[:, 36]
        df_el_appliance = df_el_ap_norm.mul(df_standard, axis=0)
    elif (status == 'aim') or (status == 'target'):  # low electricity consumption
        df_aim = df["data"].iloc[:, 37]
        df_el_appliance = df_el_ap_norm.mul(df_aim, axis=0)
    else:
        raise 'Building status unknown'

    df_el_light = df["calculs"].iloc[:, 26:50]
    df_el_light.columns = df_el_appliance.columns  # columns 1-24 - hours of the day

    df_el_add = df["calculs"].iloc[:, 52:76]  # additional light profile f.e. for "showrooms"
    df_el_add.columns = df_el_appliance.columns  # columns 1-24 - hours of the day

    df_heat_gain = df["calculs"].iloc[:, 102:126]  # heatgain by people
    df_heat_gain.columns = df_el_appliance.columns  # columns 1-24 - hours of the day

    df_dhw = df["calculs"].iloc[:, 127:151]
    df_dhw.columns = df_el_appliance.columns  # columns 1-24 - hours of the day

    df_occupancy = df["calculs"].iloc[:, 208:232]
    df_occupancy.columns = df_el_appliance.columns  # columns 1-24 - hours of the day

    return df_el_add, df_el_light, df_el_appliance, df_dhw, df_occupancy, df_heat_gain


def daily_profiles_with_monthly_deviation(status, rooms, date, df):
    # get monthly deviation
    df_months = df['profiles'].iloc[:, 49:61]

    # rename to match python timestamp
    dict_months = {'Janvier': 1, 'Fevrier': 2, 'Mars': 3, 'Avril': 4, 'Mai': 5, 'Juin': 6,
                   'Juillet': 7, 'Aout': 8, 'Septembre': 9, 'Octobre': 10, 'Novembre': 11, 'Decembre': 12}
    df_months = df_months.rename(columns=dict_months)

    monthly_factor = df_months[date.month]

    # get weekly deviation
    df_free = df['profiles'].iloc[:, 61]  # 0 no days unused, 2 = weekend unused, 1 = Sundays unused

    weekly_factor = np.repeat(1, len(rooms))
    if date.weekday() == 6:  # 6 = Sunday
        weekly_factor[(df_free == 1) | (df_free == 2)] = 0  # rooms which are not used on a Sunday
    if date.weekday() == 5:  # 5 = Saturday
        weekly_factor[df_free == 2] = 0  # rooms which are not used on a Saturday

    # get 2024 profiles
    df_el_appliance, df_el_light, df_el_add, df_dhw, df_occupancy, df_heat_gain = read_sia_2024_profiles(status, df)
    df_el = df_el_appliance + df_el_light + df_el_add  # W/m2
    df_el_gain = df_el_appliance + df_el_light
    # adjust for current day of the year
    df_el = df_el.multiply(weekly_factor * monthly_factor, axis=0)
    df_el_gain = df_el_gain.multiply(weekly_factor * monthly_factor, axis=0)
    df_dhw = df_dhw.multiply(weekly_factor * monthly_factor, axis=0)
    df_occupancy = df_occupancy.multiply(weekly_factor * monthly_factor, axis=0)
    df_heat_gain = df_heat_gain.multiply(weekly_factor * monthly_factor, axis=0)

    # get profiles for each room. multiply by usage. nan if not appearing in building
    df_el = df_el.multiply(rooms.values, axis=0).dropna()
    df_el_gain = df_el_gain.multiply(rooms.values, axis=0).dropna()
    df_dhw = df_dhw.multiply(rooms.values, axis=0).dropna()
    df_occupancy = df_occupancy.multiply(rooms.values, axis=0).dropna()
    df_heat_gain = df_heat_gain.multiply(rooms.values, axis=0).dropna()

    # aggregate
    df_profiles = pd.DataFrame()
    df_profiles['electricity_W/m2'] = df_el.sum(axis=0)
    df_profiles['hotwater_l/m2'] = df_dhw.sum(axis=0)
    df_profiles['occupancy'] = df_occupancy.sum(axis=0)
    df_profiles['elecgain_W/m2'] = df_el_gain.sum(axis=0)
    df_profiles['heatgainpeople_W/m2'] = df_heat_gain.sum(axis=0)

    return df_profiles



# All rights reserved. ECOLE POLYTECHNIQUE FEDERALE DE LAUSANNE, Switzerland,
# IPESE Laboratory, Copyright 2021
# This work can be distributed under the Apache Software License.
# See the LICENSE file for more details.

from setuptools import setup, find_packages


def read_file(file_path):
    with open(file_path, 'r', encoding='utf-8') as file:
        return file.read()


setup(

    name='REHO',
    version='1.1.0',
    packages=find_packages(),
    include_package_data=True,
    install_requires=['amplpy>=0.12.0,<0.13.0',
                      'ampl_module_base',
                      'ampl_module_highs',
                      'pandas>=1.5.3,<2.0.0',
                      'openpyxl>=3.1.2,<4.0.0',
                      'numpy>=1.23.4,<2.0.0',
                      'scipy>=1.9.2,<2.0.0',
                      'scikit-learn>=1.2.2,<2.0.0',
                      'scikit-learn-extra>=0.3.0',
                      'sqlalchemy>=1.4.42,<2.0.0',
                      'geopandas>=0.12.2,<1.0.0',
                      'matplotlib>=3.6.1,<4.0.0',
                      'plotly>=5.10,<6.0.0',
                      'kaleido>=0.2.1,<1.0.0',
                      'python-dotenv>=1.0',
                      'requests>=2.0.0,<3.0.0',
                      'coloredlogs>=15.0.1',
                      'SALib>=1.4.8',
                      'qmcpy>=1.4.5',
                      'pvlib==0.10.*'
                      ],
    package_data={
          '': ['*.csv', '*.xlsx', '*.dat', '*.txt' '*.mod', '*.ini'],
      },
    author='Dorsan Lepour',
    author_email='dorsan.lepour@epfl.ch',
    maintainer="IT team of IPESE Laboratory from EPFL",
    maintainer_email="joao.ferreiradasilva@epfl.ch",
    description='Renewable Energy Hub Optimizer (REHO) - A Comprehensive Decision Support Tool for Sustainable Energy System Planning',
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/IPESE/REHO',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.10',
    ],
    keywords='MILP, decision support, sustainable energy systems, district optimization',
    project_urls={
        'Documentation': 'https://reho.readthedocs.io/en/main/',
        'Repository': 'https://github.com/IPESE/REHO',
        'Download': 'https://pypi.org/project/REHO',
    },
)
